#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import logging
import os
import pickle
from tkinter import messagebox
from typing import Dict


class CachedDictionary:
    """ Use a Python Pickle file to maintain a cached dictionary """

    def __init__(self, cache_directory, fname):
        self.logger = logging.getLogger(__name__)
        self.cache_directory = cache_directory
        self.fname = fname
        self.dict: Dict[str, str] = {}
        self.error = False

    def get(self, val):
        return self.dict.get(val)

    def set(self, key, val):
        self.dict[key] = val

    def read(self):
        # Load Pickle file into dictionary

        if self.cache_directory is None:
            self.logger.debug(f'No directory specified for {self.fname}')
            return True
        path = os.path.join(self.cache_directory, self.fname)
        if os.path.exists(path):
            with open(path, 'rb') as file:
                self.dict = pickle.load(file)
                self.logger.debug(f'Read success CachedDict dir={self.cache_directory} fname={self.fname} len={len(self.dict)}')
                self.error = False
                return False
        else:
            self.logger.error("Missing {}".format(path))
            # Create empty file
            try:
                with open(path, 'wb') as file:
                    pickle.dump(self.dict, file)
            except OSError as e:
                messagebox.showwarning('File Error',f'{e}')
            self.error = True
            return True

    def write(self):
        # Write dictionary to Pickle file
        if self.cache_directory is None:
            return True
        path = os.path.join(self.cache_directory, self.fname)
        self.logger.debug("Write {}".format(path))

        try:
            with open(path, 'wb') as file:
                pickle.dump(self.dict, file)
        except OSError as e:
            messagebox.showwarning('File Write Error',e)
            return  True

        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import logging
import os
import re
import sys
import time
from tkinter import messagebox

from geofinder import DB, Loc, GeoKeys, MatchScore, Country
from geofinder.GeoKeys import Query, Result, Entry, get_soundex


class GeoDB:
    """
    geoname data database.  Add items, look up items, create tables, indices
    """

    def __init__(self, db_path, version):
        self.logger = logging.getLogger(__name__)
        self.start = 0
        self.match = MatchScore.MatchScore()

        self.db_path = db_path
        # See if DB exists
        if os.path.exists(db_path):
            db_exists = True
        else:
            db_exists = False

        self.db = DB.DB(db_path)
        if self.db.err:
            self.logger.error(f"Error! cannot open database {db_path}.")
            raise ValueError('Cannot open database')

        # If DB exists
        if db_exists:
            # Run sanity test on DB
            res = self.db.db_test('main.geodata')

            if res:
                self.logger.warning(f'DB error for {db_path}')
                if messagebox.askyesno('Error',
                                       f'Geoname database is empty or corrupt:\n\n {db_path} \n\nDo you want to delete it and rebuild?'):
                    messagebox.showinfo('', 'Deleting Geoname database')
                    self.db.conn.close()
                    os.remove(db_path)
                sys.exit()
        else:
            # DB didnt exist.  Create tables.
            self.create_tables()
            if version:
                self.insert_version(version)

        self.db.set_speed_pragmas()
        self.db.set_params(order_str='', limit_str='LIMIT 105')
        self.geoid_main_dict = {}  # Key is GEOID, Value is DB ID for entry
        self.geoid_admin_dict = {}  # Key is GEOID, Value is DB ID for entry
        self.place_type = ''

    def delete_dbZZZ(self):
        self.logger.info('Deleting geoname DB')

    def lookup_place(self, place: Loc.Loc) -> []:
        """
        Lookup a place in our geoname.org dictionary and update place with Geo_result with lat, long, District, etc
        The dictionary geo_result entry contains: Lat, Long, districtID (County or State or Province ID)
        There can be multiple entries if a city name isnt unique in a country
        """
        result_place: Loc = Loc.Loc()
        self.start = time.time()
        place.result_type = Result.STRONG_MATCH
        #place.admin2_name, modified = GeoKeys.admin2_normalize(place.admin2_name, place.country_iso)

        if place.country_iso != '' and place.country_name == '':
            place.country_name = self.get_country_name(place.country_iso)

        # Lookup Place based on Place Type
        if place.place_type == Loc.PlaceType.ADMIN1:
            self.select_admin1(place)
        elif place.place_type == Loc.PlaceType.ADMIN2:
            if place.admin1_id == '':
                self.get_admin1_id(place=place)
            self.select_admin2(place)
            #if len(place.georow_list) == 0:
                # Try search with some text replacements
                #place.admin2_name, modified = GeoKeys.admin2_normalize(place.admin2_name, place.country_iso)
                #if modified:
                #    self.select_admin2(place)
        elif place.place_type == Loc.PlaceType.COUNTRY:
            self.select_country(place)
        elif place.place_type == Loc.PlaceType.ADVANCED_SEARCH:
            self.advanced_search(place)
        else:
            # Lookup as City
            if place.admin2_id == '':
                self.get_admin2_id(place=place)
            self.select_city(place)

        # nm = place.original_entry
        # self.logger.debug(f'Search results for {place.target} pref[{place.prefix}]')
        min_score = 9999

        # Add search quality score to each entry
        for idx, rw in enumerate(place.georow_list):
            self.copy_georow_to_place(row=rw, place=result_place)

            if len(place.prefix) > 0 and result_place.prefix == '':
                result_place.prefix = ' '
                result_place.prefix_commas = ','
            else:
                result_place.prefix = ''

            score = self.match.match_score(inp_place=place, res_place=result_place)
            if score < min_score:
                min_score = score

            # Convert row tuple to list and extend so we can assign score
            update = list(rw)
            update.append(1)
            update[GeoKeys.Entry.SCORE] = score
            place.georow_list[idx] = tuple(update)  # Convert back from list to tuple

            # Remove items in prefix that are in result
            tk_list = result_place.original_entry.split(",")
            if place.place_type != Loc.PlaceType.ADVANCED_SEARCH:
                for item in tk_list:
                    place.prefix = re.sub(item.strip(' ').lower(), '', place.prefix)

        if place.result_type == Result.STRONG_MATCH and len(place.prefix) > 0:
            place.result_type = Result.PARTIAL_MATCH

        if place.result_type == Result.STRONG_MATCH and min_score > 10:
            place.result_type = Result.PARTIAL_MATCH

    def select_city(self, place: Loc):
        """
        Search for  entry - try the most exact match first, then less exact matches
        """
        lookup_target = place.target
        if len(lookup_target) == 0:
            return
        pattern = self.create_wildcard(lookup_target)
        quick_pattern = self.create_quick_wildcard(lookup_target)

        sdx = get_soundex(lookup_target)
        #self.logger.debug(f'CITY lkp targ=[{lookup_target}] adm1 id=[{place.admin1_id}]'
        #                  f' adm2 id=[{place.admin2_id}] iso=[{place.country_iso}] patt =[{pattern}] sdx={sdx} pref={place.prefix}')

        query_list = []

        if len(place.country_iso) == 0:
            # No country present - try lookup by name.
            query_list.append(Query(where="name = ?",
                                    args=(lookup_target,),
                                    result=Result.PARTIAL_MATCH))
            # lookup by wildcard name
            query_list.append(Query(where="name LIKE ?",
                                    args=(quick_pattern,),
                                    result=Result.WILDCARD_MATCH))
            # lookup by soundex
            query_list.append(Query(where="sdx = ?",
                                    args=(sdx,),
                                    result=Result.SOUNDEX_MATCH))

            place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)
            # self.logger.debug(place.georow_list)
            return

        # Build query list - try each query in order until a match is found
        # Start with the most exact match depending on the data provided.
        if len(place.admin1_name) > 0:
            # lookup by name, ADMIN1, country
            query_list.append(Query(where="name = ? AND country = ? AND admin1_id = ?",
                                    args=(lookup_target, place.country_iso, place.admin1_id),
                                    result=Result.STRONG_MATCH))

            # lookup by wildcard name, ADMIN1, country
            query_list.append(Query(where="name LIKE ? AND country = ? AND admin1_id = ?",
                                    args=(pattern, place.country_iso, place.admin1_id),
                                    result=Result.WILDCARD_MATCH))
        else:
            # lookup by wildcard  name, country
            query_list.append(Query(where="name LIKE ? AND country = ?",
                                    args=(pattern, place.country_iso),
                                    result=Result.WILDCARD_MATCH))

        # Lookup by name, country
        query_list.append(Query(where="name = ? AND country = ?",
                                args=(lookup_target, place.country_iso),
                                result=Result.PARTIAL_MATCH))

        if len(place.admin1_name) > 0:
            # lookup by Soundex name, country and admin1
            query_list.append(Query(where="sdx = ? AND admin1_id = ? AND country = ?",
                                    args=(sdx, place.admin1_id, place.country_iso),
                                    result=Result.SOUNDEX_MATCH))
        else:
            # lookup by Soundex name, country
            query_list.append(Query(where="sdx = ? AND country = ?",
                                    args=(sdx, place.country_iso),
                                    result=Result.SOUNDEX_MATCH))

        # Try each query in list
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata',
                                                                          query_list=query_list)

    def select_admin2(self, place: Loc):
        """Search for Admin2 entry"""
        lookup_target = place.admin2_name
        if len(lookup_target) == 0:
            return
        place.target = lookup_target
        # sdx = get_soundex(lookup_target)

        # Try Admin query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND country = ? AND admin1_id = ? AND f_code=?",
                  args=(lookup_target, place.country_iso, place.admin1_id, 'ADM2'),
                  result=Result.STRONG_MATCH),
            Query(where="name = ? AND country = ? AND f_code=?",
                  args=(lookup_target, place.country_iso, 'ADM2'),
                  result=Result.PARTIAL_MATCH),
            Query(where="name LIKE ? AND country = ? AND f_code=?",
                  args=(self.create_wildcard(lookup_target), place.country_iso, 'ADM2'),
                  result=Result.PARTIAL_MATCH),
            Query(where="name = ?  AND f_code=?",
                  args=(lookup_target, 'ADM2'),
                  result=Result.PARTIAL_MATCH),
            Query(where="name LIKE ? AND country = ? AND f_code=?",
                  args=(self.create_county_wildcard(lookup_target), place.country_iso, 'ADM2'),
                  result=Result.WILDCARD_MATCH)
        ]

        # self.logger.debug(f'Admin2 lookup=[{lookup_target}] country=[{place.country_iso}]')
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)
        if place.result_type == GeoKeys.Result.WILDCARD_MATCH:
            # Found as Admin2 without shire
            place.original_entry = re.sub('shire', '', place.original_entry)

        if len(place.georow_list) == 0:
            # Try city rather than County match.
            save_admin2 = place.admin2_name
            place.city1 = place.admin2_name
            place.admin2_name = ''
            # self.logger.debug(f'Try admin2 as city: [{place.target}]')

            self.select_city(place)

            if len(place.georow_list) == 0:
                #  not found.  restore admin
                place.admin2_name = save_admin2
                place.city1 = ''
            else:
                # Found match as a City
                place.place_type = Loc.PlaceType.CITY
                match_adm1 = self.get_admin1_name_direct(lookup_target=place.georow_list[0][Entry.ADM1], iso=place.country_iso)
                # self.logger.debug(f'pl_iso [{place.country_iso}] pl_adm1 {place.admin1_name} match_adm1=[{match_adm1}] ')
                if place.admin1_name != match_adm1:
                    place.prefix = place.admin1_name.title()
                    place.admin1_name = ''
                return

    def select_admin1(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.admin1_name

        pattern = self.create_wildcard(lookup_target)
        if len(lookup_target) == 0:
            return
        sdx = get_soundex(lookup_target)

        # self.logger.debug(f'sel adm1 patt={pattern} iso={place.country_iso}')

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND country = ? AND f_code = ? ",
                  args=(lookup_target, place.country_iso, 'ADM1'),
                  result=Result.STRONG_MATCH),
            Query(where="name LIKE ? AND country = ?  AND f_code = ?",
                  args=(pattern, place.country_iso, 'ADM1'),
                  result=Result.WILDCARD_MATCH),
            Query(where="sdx = ? AND country = ? AND f_code=?",
                  args=(sdx, place.country_iso, 'ADM1'),
                  result=Result.SOUNDEX_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

    def select_country(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.country_iso
        if len(lookup_target) == 0:
            return
        sdx = get_soundex(lookup_target)

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="country = ? AND f_code = ? ",
                  args=(place.country_iso, 'ADM0'),
                  result=Result.STRONG_MATCH),
            Query(where="sdx = ?  AND f_code=?",
                  args=(sdx, 'ADM0'),
                  result=Result.SOUNDEX_MATCH)
        ]

        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)
        # self.build_result_list(place.georow_list, place.event_year)

    def get_admin1_id(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.admin1_name
        if len(lookup_target) == 0:
            return

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND country = ? AND f_code = ? ",
                  args=(lookup_target, place.country_iso, 'ADM1'),
                  result=Result.STRONG_MATCH),
            Query(where="name LIKE ? AND country = ?  AND f_code = ?",
                  args=(self.create_wildcard(lookup_target), place.country_iso, 'ADM1'),
                  result=Result.WILDCARD_MATCH),
            Query(where="name = ?  AND f_code = ?",
                  args=(lookup_target, 'ADM1'),
                  result=Result.SOUNDEX_MATCH)]

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            place.admin1_id = row_list[0][Entry.ADM1]
            # Fill in Country ISO
            if place.country_iso == '':
                place.country_iso = row_list[0][Entry.ISO]

    def get_admin2_id(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.admin2_name
        if len(lookup_target) == 0:
            return

        # Try each query until we find a match - each query gets less exact
        query_list = []
        if len(place.admin1_id) > 0:
            query_list.append(Query(where="name = ? AND country = ? AND admin1_id=? AND f_code=?",
                                    args=(lookup_target, place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.STRONG_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? and admin1_id = ? AND f_code=?",
                                    args=(self.create_wildcard(lookup_target), place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? and admin1_id = ? AND f_code=?",
                                    args=(self.create_county_wildcard(lookup_target), place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
        else:
            query_list.append(Query(where="name = ? AND country = ? AND f_code=?",
                                    args=(lookup_target, place.country_iso, 'ADM2'),
                                    result=Result.STRONG_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? AND f_code=?",
                                    args=(self.create_wildcard(lookup_target), place.country_iso, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? AND f_code=?",
                                    args=(self.create_county_wildcard(lookup_target), place.country_iso, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            place.admin2_id = row[Entry.ADM2]

    def get_admin1_alt_name(self, place: Loc) -> (str, str):
        """Search for Admin1 entry"""
        lookup_target = place.admin1_id
        if len(lookup_target) == 0:
            return '', ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin1_id = ? AND country = ?  AND f_code = ? ",
                  args=(lookup_target, place.country_iso, 'ADM1'),
                  result=Result.STRONG_MATCH)
        ]
        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            admin1_name, lang = self.get_alt_name(row[Entry.ID])
            return admin1_name, lang
        else:
            return '', ''

    def get_admin1_name(self, place: Loc) -> str:
        """Search for Admin1 entry"""
        lookup_target = place.admin1_id
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin1_id = ? AND country = ?  AND f_code = ? ",
                  args=(lookup_target, place.country_iso, 'ADM1'),
                  result=Result.STRONG_MATCH)
        ]
        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            place.admin1_name = row[Entry.NAME]
            return place.admin1_name
        else:
            return ''

    def get_admin1_name_direct(self, lookup_target, iso) -> str:
        """Search for Admin1 entry"""
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin1_id = ? AND country = ?  AND f_code = ? ",
                  args=(lookup_target, iso, 'ADM1'),
                  result=Result.STRONG_MATCH)
        ]
        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            return row[Entry.NAME]
        else:
            return ''

    def get_admin2_name_direct(self, admin1_id, admin2_id, iso) -> str:
        """Search for Admin2 entry"""
        lookup_target = admin2_id
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin2_id = ? AND country = ? AND admin1_id = ?",
                  args=(lookup_target, iso, admin1_id),
                  result=Result.STRONG_MATCH),
            Query(where="admin2_id = ? AND country = ?",
                  args=(lookup_target, iso),
                  result=Result.PARTIAL_MATCH)]

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            return row[Entry.NAME]
        else:
            return ''

    def get_admin2_name(self, place: Loc) -> str:
        """Search for Admin1 entry"""
        lookup_target = place.admin2_id
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin2_id = ? AND country = ? AND admin1_id = ?",
                  args=(lookup_target, place.country_iso, place.admin1_id),
                  result=Result.STRONG_MATCH),
            Query(where="admin2_id = ? AND country = ?",
                  args=(lookup_target, place.country_iso),
                  result=Result.PARTIAL_MATCH)]

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            place.admin2_name = row[Entry.NAME]
            # self.logger.debug(f'adm2 nm = {place.admin2_name}')
            return place.admin2_name
        else:
            return ''

    def lookup_geoid(self, place: Loc) -> None:
        """Search for GEOID"""
        result_place: Loc = Loc.Loc()

        query_list = [
            Query(where="geoid = ? ",
                  args=(place.target,),
                  result=Result.STRONG_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)
        if len(place.georow_list) == 0:
            place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)
        else:
            place.georow_list = place.georow_list[:1]
            place.result_type = GeoKeys.Result.STRONG_MATCH

        # Add search quality score to each entry
        for idx, rw in enumerate(place.georow_list):
            self.copy_georow_to_place(row=rw, place=result_place)
            update = list(rw)
            update.append(1)  # Extend list row and assign score
            result_place.prefix = ''
            res_nm = result_place.format_full_nm(None)
            score = 0.0

            # Remove items in prefix that are in result
            tk_list = res_nm.split(",")
            for item in tk_list:
                place.prefix = re.sub(item.strip(' ').lower(), '', place.prefix)

            update[GeoKeys.Entry.SCORE] = int(score * 100)
            place.georow_list[idx] = tuple(update)

    def lookup_main_dbid(self, place: Loc) -> None:
        """Search for DB ID"""
        query_list = [
            Query(where="id = ? ",
                  args=(place.target,),
                  result=Result.STRONG_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)

    def lookup_admin_dbid(self, place: Loc) -> None:
        """Search for DB ID"""
        query_list = [
            Query(where="id = ? ",
                  args=(place.target,),
                  result=Result.STRONG_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

    def get_country_name(self, iso: str) -> str:
        """ return country name for specified ISO code """
        if len(iso) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="country = ? AND f_code = ? ",
                  args=(iso, 'ADM0'),
                  result=Result.STRONG_MATCH)]

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            res = row_list[0][Entry.NAME]
            if iso == 'us':
                res = 'United States'
        else:
            res = ''
        return res

    def get_country_iso(self, place: Loc) -> str:
        """ Return ISO code for specified country"""
        lookup_target, modified = GeoKeys.country_normalize(place.country_name)
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND f_code = ? ",
                  args=(lookup_target, 'ADM0'),
                  result=Result.STRONG_MATCH),
            # Query(where="name LIKE ?  AND f_code = ? ",
            #      args=(self.create_wildcard(lookup_target), 'ADM0'),
            #      result=Result.PARTIAL_MATCH)  #,
            # Query(where="sdx = ?  AND f_code = ? ",
            #      args=(GeoKeys.get_soundex (lookup_target), 'ADM0'),
            #      result=Result.PARTIAL_MATCH)
        ]

        row_list, result_code = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            res = row_list[0][Entry.ISO]
            if len(row_list) == 1:
                place.country_name = row_list[0][Entry.NAME]
        else:
            res = ''

        return res


    def advanced_search(self, place: Loc):
        """
        Advanced search - support parameters for ISO and Feature class
        """
        lookup_target = place.target
        if len(lookup_target) == 0:
            return
        pattern = self.create_wildcard(lookup_target)
        feature_pattern = self.create_wildcard(place.feature)
        self.logger.debug(f'Advanced Search. Targ=[{pattern}] feature=[{feature_pattern}]'
                          f'  iso=[{place.country_iso}] ')

        if len(place.feature) > 0:
            query_list = [
                Query(where="name LIKE ? AND country LIKE ? AND f_code LIKE ?",
                      args=(pattern, place.country_iso, feature_pattern),
                      result=Result.PARTIAL_MATCH)]
        else:
            query_list = [
                Query(where="name LIKE ? AND country LIKE ?",
                      args=(pattern, place.country_iso),
                      result=Result.PARTIAL_MATCH)]

        # Search main DB
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)

        # self.logger.debug(f'main Result {place.georow_list}')

        # Search admin DB
        admin_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)
        place.georow_list.extend(admin_list)
        # self.logger.debug(f'admin Result {place.georow_list}')

    def copy_georow_to_place(self, row, place: Loc):
        # Copy data from DB row into Place
        # self.logger.debug(row)
        place.admin1_id = ''
        place.admin2_id = ''
        place.city1 = ''

        place.country_iso = str(row[Entry.ISO])
        place.country_name = str(self.get_country_name(row[Entry.ISO]))
        place.lat = row[Entry.LAT]
        place.lon = row[Entry.LON]
        place.feature = str(row[Entry.FEAT])
        place.geoid = str(row[Entry.ID])

        if place.feature == 'ADM0':
            self.place_type = Loc.PlaceType.COUNTRY
            pass
        elif place.feature == 'ADM1':
            place.admin1_id = row[Entry.ADM1]
            self.place_type = Loc.PlaceType.ADMIN1
        elif place.feature == 'ADM2':
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            self.place_type = Loc.PlaceType.ADMIN2
        else:
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            place.city1 = row[Entry.NAME]
            self.place_type = Loc.PlaceType.CITY

        place.admin1_name = str(self.get_admin1_name(place))
        place.admin2_name = str(self.get_admin2_name(place))
        if place.admin2_name is None:
            place.admin2_name = ''
        if place.admin1_name is None:
            place.admin1_name = ''

        place.city1 = str(place.city1)
        if place.city1 is None:
            place.city1 = ''

    def clear_geoname_data(self):
        # Delete all the geoname data
        for tbl in ['geodata', 'admin']:
            # noinspection SqlWithoutWhere
            self.db.delete_table(tbl)

    # Streams_nocase_idx ON Streams(Name 

    def create_geoid_index(self):
        # Create indices
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS geoid_idx ON geodata(geoid )')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS admgeoid_idx ON admin(geoid  )')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS altnamegeoid_idx ON altname(geoid  )')

    def create_indices(self):
        # Create indices
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS name_idx ON geodata(name, country )')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS admin1_idx ON geodata(admin1_id )')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS sdx_idx ON geodata(sdx )')

        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS adm_name_idx ON admin(name, country )')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS adm_admin1_idx ON admin(admin1_id, f_code)')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS adm_admin2_idx ON admin(admin1_id, admin2_id)')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS adm_country_idx ON admin(country, f_code)')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS adm_sdx_idx ON admin(sdx )')

    @staticmethod
    def make_georow(name: str, iso: str, adm1: str, adm2: str, lat: float, lon: float, feat: str, geoid: str, sdx: str) -> ():
        # Return a dummy geo-row
        res = (name, iso, adm1, adm2, lat, lon, feat, geoid, sdx)
        return res

    def get_row_count(self) -> int:
        return self.db.get_row_count()

    @staticmethod
    def create_wildcard(pattern):
        # Create SQL wildcard pattern (convert * to %).  Add % on end
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'%{pattern}%'

    @staticmethod
    def create_quick_wildcard(pattern):
        # Create SQL wildcard pattern (convert * to %).  Add % on end
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'{pattern}%'

    @staticmethod
    def create_county_wildcard(pattern):
        # Try pattern with 'shire' removed
        pattern = re.sub(r"shire", "", pattern)
        # Create SQL wildcard pattern (convert * to %).  Add % on end
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'{pattern}'

    def close(self):
        self.db.set_analyze_pragma()
        self.logger.info('Closing Database')
        self.db.conn.close()

    def set_display_names(self, temp_place):
        place_lang = Country.Country.get_lang(temp_place.country_iso)
        res, lang = self.get_alt_name(temp_place.geoid)
        if res != '' and (lang == place_lang or lang == 'ut8'):
            temp_place.city1 = res

        res, lang = self.get_admin1_alt_name(temp_place)
        if res != '' and (lang == place_lang or lang == 'ut8'):
            temp_place.admin1_name = res

    def get_alt_name(self, geoid) -> (str, str):
        # retrieve alternate name
        query_list = [
            Query(where="geoid = ?",
                  args=(geoid,),
                  result=Result.STRONG_MATCH)]
        select = 'name, lang'
        row_list, res = self.db.process_query(select_string=select, from_tbl='main.altname', query_list=query_list)
        if len(row_list) > 0:
            return row_list[0][0], row_list[0][1]
        else:
            return '', ''

    def insert(self, geo_row: (), feat_code: str):
        # We split the data into 2  tables, 1) Admin: ADM0/ADM1/ADM2,  and 2) city data
        if feat_code == 'ADM1' or feat_code == 'ADM0' or feat_code == 'ADM2':
            sql = ''' INSERT OR IGNORE INTO admin(name,country, admin1_id,admin2_id,lat,lon,f_code, geoid, sdx)
                      VALUES(?,?,?,?,?,?,?,?,?) '''
            row_id = self.db.execute(sql, geo_row)
            # Add name to dictionary.  Used by AlternateNames for fast lookup during DB build
            self.geoid_admin_dict[geo_row[Entry.ID]] = row_id
        else:
            sql = ''' INSERT OR IGNORE INTO geodata(name, country, admin1_id, admin2_id, lat, lon, f_code, geoid, sdx)
                      VALUES(?,?,?,?,?,?,?,?,?) '''
            row_id = self.db.execute(sql, geo_row)
            # Add name to dictionary.  Used by AlternateNames for fast lookup during DB build
            self.geoid_main_dict[geo_row[Entry.ID]] = row_id

        return row_id

    def insert_alternate_name(self, alternate_name: str, geoid: str, lang: str):
        # We split the data into 2  tables, 1) Admin: ADM0/ADM1/ADM2,  and 2) city data
        row = (alternate_name, lang, geoid)
        sql = ''' INSERT OR IGNORE INTO altname(name,lang, geoid)
                  VALUES(?,?,?) '''
        row_id = self.db.execute(sql, row)

    def insert_version(self, db_version: int):
        self.db.begin()

        sql = ''' INSERT OR IGNORE INTO version(version)
                  VALUES(?) '''
        args = (db_version,)
        row_id = self.db.execute(sql, args)
        self.db.commit()

    def get_db_version(self) -> int:
        # If version table does not exist, this is V1
        if self.db.table_exists('version'):
            # query version ID
            query_list = [
                Query(where="version like ?",
                      args=('%',),
                      result=Result.STRONG_MATCH)]
            select_str = '*'
            row_list, res = self.db.process_query(select_string=select_str, from_tbl='main.version', query_list=query_list)
            if len(row_list) > 0:
                ver = int(row_list[0][1])
                self.logger.debug(f'Database Version = {ver}')
                return ver

        # No version table, so this is V1
        self.logger.debug('No version table.  Version is 1')
        return 1

    def create_tables(self):
        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_geodata_table = """CREATE TABLE IF NOT EXISTS geodata    (
                id           integer primary key autoincrement not null,
                name     text,
                country     text,
                admin1_id     text,
                admin2_id text,
                lat      text,
                lon       text,
                f_code      text,
                geoid      text,
                sdx     text
                                    );"""

        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_admin_table = """CREATE TABLE IF NOT EXISTS admin    (
                id           integer primary key autoincrement not null,
                name     text,
                country     text,
                admin1_id     text,
                admin2_id text,
                lat      text,
                lon       text,
                f_code      text,
                geoid      text,
                sdx     text
                                    );"""

        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_alt_name_table = """CREATE TABLE IF NOT EXISTS altname    (
                id           integer primary key autoincrement not null,
                name     text,
                lang     text,
                geoid      text
                                    );"""

        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_version_table = """CREATE TABLE IF NOT EXISTS version    (
                id           integer primary key autoincrement not null,
                version     integer
                                    );"""

        for tbl in [sql_geodata_table, sql_admin_table, sql_version_table, sql_alt_name_table]:
            self.db.create_table(tbl)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import configparser
import os
import sys
from pathlib import Path
from tkinter import messagebox, filedialog

from geofinder import GeoKeys


class IniHandler:
    def __init__(self, home_path:str, ini_name):
        self.directory:Path = Path()
        #self.home_path = str(Path.home())
        self.home_path:Path = Path(home_path)

        self.ini = configparser.ConfigParser()
        self.ini_path:Path = Path(os.path.join(str(self.home_path), ini_name))

    def ini_read(self, section, key):
        # read item from ini file
        val = None
        try:
            self.ini.read(self.ini_path)
            val = self.ini.get(section, key)
        except (configparser.NoSectionError, configparser.NoOptionError, configparser.MissingSectionHeaderError) as e:
            print ('no section')

        # read value from a section
        return val

    def ini_set(self, section, key: str, val: str):
        # update existing value
        self.ini.set(section, key, val)
        # save to a file
        with open(str(self.ini_path), 'w') as configfile:
            self.ini.write(configfile)

    def ini_add_section(self, section):
        # add a new section and some values
        try:
            self.ini.add_section(section)
        except configparser.DuplicateSectionError:
            print('Duplicate section')

        # save to a file
        with open(str(self.ini_path), 'w') as configfile:
            self.ini.write(configfile)

    def get_directory_from_ini(self) -> str:
        if self.ini_path.is_file():
            val = self.ini_read('PATH', 'DIRECTORY')
            if val:
                self.directory = val
            else:
                # Not Found.  Create INI file
                self.directory = Path(os.path.join(str(self.home_path), GeoKeys.get_directory_name()))
                self.ini_add_section('PATH')
                self.ini_set(section='PATH', key='DIRECTORY', val=str(self.directory))
        else:
            # Not Found.  Create INI file
            self.directory = Path(os.path.join(str(self.home_path), GeoKeys.get_directory_name()))
            self.ini_add_section('PATH')
            self.ini_set(section='PATH', key='DIRECTORY', val=str(self.directory))

        #  if directory doesnt exist, prompt user for folder
        if not Path(self.directory).is_dir():
            messagebox.showinfo('Geofinder Folder not found', 'Choose Folder for GeoFinder data in next dialog')
            self.directory = filedialog.askdirectory(initialdir=self.home_path, title="Choose Folder for GeoFinder data")
            if len(self.directory) == 0:
                sys.exit()
            else:
                self.ini_add_section('PATH')
                self.ini_set(section='PATH', key='DIRECTORY', val=str(self.directory))
        return self.directory



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import os
import re
import xml.etree.ElementTree as Tree
from io import BytesIO
from typing import Union

from geofinder import Progress, GrampsCsv
from geofinder.AncestryFile import AncestryFile


# 0Place (ID), 1Title, 2Name, 3Type, 4latitude, 5longitude, enclosed_by


# State Machine
class State:
    PASS_THROUGH = 0
    COLLECT_PLACE_XML = 1
    WALK_PLACE_TREE = 2


class GrampsXml(AncestryFile):
    """
    Parse and update a Gramps XML DTD 1.7.1 file - sample of Place section below
    Gramps XML (Family Tree) - Not compressed, Not package

    https://www.gramps-project.org/xml/

    <places>
    <placeobj handle="_e3e603cc22486b8f5a7b6dda58" change="1571551969" id="P0565" type="County">
      <ptitle>Middlesex County, Massachusetts, United States</ptitle>
      <pname value="Middlesex County"/>
      <coord long="-71.39184" lat="42.48555"/>
      <placeref hlink="_e3e603bd7c12a78d92ba44844fd"/>
    </placeobj>

    <placeobj handle="_e3e603bc1c94b85f1b59026fa08" change="1566106543" id="P0001" type="Unknown">
      <ptitle>Palo Alto, Santa Clara County, California, United States</ptitle>
      <pname value="Palo Alto, Santa Clara County, California, United States"/>
      <coord long="-122.14302" lat="37.44188"/>
    </placeobj>
    </places>

    xmllint --c14n one.xml > 1.xml
    """

    def __init__(self, in_path: str, out_suffix: str, cache_d, progress: Union[None, Progress.Progress], geodata):
        super().__init__(in_path, out_suffix, cache_d, progress, geodata)
        self.xml_tree = None
        self.collect_lines = False
        self.place_xml_lines = b''
        self.place = None
        self.child = None
        self.state = State.PASS_THROUGH  # Write out each line as-is unless we are in Place section
        self.got_pname = False
        self.got_place = False
        self.lon = 99.9
        self.lat = 99.9
        self.place_complete = 0
        self.csv = GrampsCsv.GrampsCsv(in_path=in_path, geodata=geodata)
        self.title = ''

    def parse_line(self, line: str):
        # Called by read_and_parse_line for each line in file
        # Accumulate all XML place entries into binary string, then build XML tree, then search XML tree
        # and return each entry in self.value with self.tag set to PLAC

        # Set State
        if '<places>' in line:
            # Reached the start of Places XML section.  Accumulate XML text until end of section
            self.state = State.COLLECT_PLACE_XML
            self.logger.debug('XML places section - start')
        elif '</places>' in line:
            # Reached the end of Places section
            # TODO - Handle case where there is additional data on </places> line, such as '</places> <objects>'
            line += '\n'
            self.place_xml_lines += bytes(line, "utf8")
            self.logger.debug(f'XML Places section - complete.  len={len(self.place_xml_lines)}')

            # Build tree from XML string
            try:
                self.xml_tree = Tree.parse(BytesIO(self.place_xml_lines))
            except TypeError:
                self.logger.warning(f'XML parse error')
                self.xml_tree = None
            self.place_total = sum([1 for entry in self.xml_tree.getiterator('placeobject')])
            self.logger.info(f'XML Parse complete. PLACE COUNT={self.place_total}')

            self.state = State.WALK_PLACE_TREE
            self.more_available = True

        # Handle line based on State
        self.tag = 'OTHER'

        if self.state == State.COLLECT_PLACE_XML:
            # Collect lines
            # Convert all placeobj tags to placeobject tag
            # As each is processed we convert it back to placeobj
            # This allows us to keep track of which Place Objects we have processed.
            #if 'placeobj' in line:
            #    self.place_total += 1
            line = re.sub('placeobj', 'placeobject', line)
            self.place_xml_lines += bytes(line, "utf8")
            #self.logger.debug(f'Collect XML [{line}]')
            self.tag = 'IGNORE'
        elif self.state == State.PASS_THROUGH:
            #  output line in get_next_place
            #self.logger.debug(f'Pass through XML [{line}]')
            pass
        elif self.state == State.WALK_PLACE_TREE:
            # Set self.value with next place
            self.find_xml_place()

            if not self.more_available:
                # Got to END OF TREE.  WRITE XML tree
                self.logger.debug('End of XML tree')
                self.write_out_tree()
                self.csv.complete_csv()

        return self.id

    def write_out_tree(self):
        # Write out XML tree
        tmp_name = self.out_path + '.tmp'
        self.logger.debug(f'Writing XML tree to tempfile: {tmp_name}')
        self.xml_tree.write(tmp_name)

        # Append XML tmp file to our output file
        self.append_file(tmp_name)

        # All additional text is pass through (not part of Place section)
        self.state = State.PASS_THROUGH
        #self.outfile.close()

    def append_file(self, temp_path):
        # Append temp file to our output
        self.outfile.flush()
        tmp_file = open(temp_path, 'r')
        # Read in chunks
        while True:
            data = tmp_file.read(65536)
            if data:
                self.outfile.write(data)
            else:
                break
        tmp_file.close()
        self.logger.debug(f'Removing temp file: {temp_path}')
        os.remove(temp_path)
        self.outfile.write('\n')
        self.outfile.flush()

    def find_xml_place(self):
        # Find the next placeobject entry in Tree
        # Each time we process a place, change XML tag to placeobj and
        # set self.tag to PLAC so we don't handle again

        # Find first placeobject
        self.plac = self.xml_tree.find('placeobject')

        if self.plac is not None:
            #self.logger.debug(f'\n\nPLACEOBJECT {self.place.tag} =========')
            self.id = self.plac.get("id")
            self.title = ''
            self.name = ''
            self.place_complete += 1
            # update progress bar
            self.percent_complete = int(self.place_complete * 100 / self.place_total)
            self.progress(f" ", self.percent_complete)

            # Walk thru each entry in place object
            for place_entry in self.plac.iter():
                self.child = place_entry
                # print(f'tag ={place_entry.tag}')
                if place_entry.tag == 'ptitle' and self.got_place is False:
                    self.tag = 'PLAC'
                    self.value = place_entry.text
                    self.title = self.value
                    #self.logger.debug(f'PTITLE <{place_entry.tag} VALUE="{self.value}"/>')
                    self.got_place = True
                    #return
                elif place_entry.tag == 'pname' and self.got_pname is False:
                    # <pname value="Chelsea, Greater London, England, United Kingdom"/>
                    self.tag = 'PLAC'
                    self.value = place_entry.get('value')
                    self.name = self.value
                    #self.logger.debug(f'PNAME <{place_entry.tag} VALUE="{self.value}"/>')
                    self.got_pname = True
                    #return
                elif place_entry.tag == 'coord':
                    # <coord long="-0.16936" lat="51.48755"/>
                    self.lon = place_entry.attrib.get('long')
                    self.lat = place_entry.attrib.get('lat')
                    self.tag = 'IGNORE'
                    #self.logger.debug(f'<{place_entry.tag} LONG="{self.lon}" LAT="{self.lat}"/>')
                    self.plac.remove(place_entry)
                else:
                    # print(f'DD <{place_entry.tag} {place_entry.attrib}>')
                    pass

            # Done walking through place element
            if self.lat != 99.9:
                # Create Coord Latitude/Longitude section
                coord_elem = Tree.Element("coord")
                coord_elem.set('long', str(self.lon))
                coord_elem.set('lat', str(self.lat))
                self.plac.append(coord_elem)

            # Change tag back to 'placeobj' so we don't visit it next time we are called
            self.plac.tag = 'placeobj'

            # Reset values
            self.got_pname = False
            self.got_place = False
        else:
            # Tree completed.  No more place objects available
            self.logger.debug('XML tree complete')
            self.more_available = False

    def write_updated(self, txt, place):
        # Update place entry in tree.  Tree will be written out later when entire XML tree is written out
        self.csv.create_csv_node(place)
        if self.child.text is not None:
            self.child.text = txt.strip(', ')
        else:
            if self.child.tag == 'pname':
                self.child.set('value', txt.strip(', '))

    def write_asis(self, entry):
        # Do nothing - No change to place entry
        # It will be written out later when entire XML tree is written out
        # TODO implement CSV
        self.csv.write_asis(entry)
        #self.logger.debug(f'AS IS {self.id} {self.value}')
        pass

    def write_lat_lon(self, lat: float, lon: float):
        """ Create an XML lat/long coordinate enty """
        if self.output_latlon is False:
            return
        self.lat = lat
        self.lon = lon



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA


import logging
import os
import time
import unittest
from pathlib import Path

from geofinder import Loc, MatchScore, Geodata


class TestScoring(unittest.TestCase):
    scoring = None
    logger = None
    geodata = None
    test_idx = -1

    # ===== TEST SCORING
    PERFECT = -20
    VERY_STRONG = -1
    STRONG = 18
    GOOD = 23
    POOR = 38
    VERY_POOR = 70
    NO_MATCH = 100.1

    test_values = [
        # Target, Result, Feature, Expected Score
        ("palo alto, santa clara, california, usa", "palo alto, santa clara, california, usa", 'PPL', STRONG),
        ("city of palo alto, santa clara, california, usa", "palo alto, santa clara, california, usa", 'ADM3', STRONG)
    ]

    @classmethod
    def setUpClass(cls):
        TestScoring.logger = logging.getLogger(__name__)
        fmt = "%(levelname)s %(name)s.%(funcName)s %(lineno)d: %(message)s"
        logging.basicConfig(level=logging.DEBUG, format=fmt)
        TestScoring.logger.debug('Scoring')
        TestScoring.scoring = MatchScore.MatchScore()

        # Load test data
        directory = os.path.join(str(Path.home()), "geoname_test")
        TestScoring.geodata = Geodata.Geodata(directory_name=directory, progress_bar=None)
        error: bool = TestScoring.geodata.read()
        if error:
            TestScoring.logger.error("Missing geodata support Files.")
            raise ValueError('Cannot open database')

        # Read in Geoname Gazeteer file - city names, lat/long, etc.
        start_time = time.time()
        error = TestScoring.geodata.read_geonames()
        end_time = time.time()

    def setUp(self) -> None:
        TestScoring.in_place: Loc.Loc = Loc.Loc()
        TestScoring.out_place: Loc.Loc = Loc.Loc()
        TestScoring.test_idx += 1

    def run_test1(self, title: str, inp, out):
        print("*****TEST: WORD {}".format(title))
        out, inp = TestScoring.scoring.remove_matching_sequences(out, inp)
        return out, inp

    @staticmethod
    def run_test2(title: str, inp, out):
        print("*****TEST: CHAR {}".format(title))
        out, inp = TestScoring.scoring.remove_matching_sequences(out, inp)
        return out, inp

    @staticmethod
    def run_test3()->int:
        """
        :param title:
        :param inp:
        :param res:
        :param feat:
        :return:
        """
        title = TestScoring.test_values[TestScoring.test_idx][0]
        inp = TestScoring.test_values[TestScoring.test_idx][0]
        res = TestScoring.test_values[TestScoring.test_idx][1]
        feat = TestScoring.test_values[TestScoring.test_idx][2]

        in_place = Loc.Loc()
        in_place.original_entry = inp
        in_place.parse_place(inp, geo_files=TestScoring.geodata.geo_files)
        if in_place.country_name == '' and in_place.country_iso != '':
            in_place.country_name = TestScoring.geodata.geo_files.geodb.get_country_name(in_place.country_iso)

        res_place = Loc.Loc()
        res_place.original_entry = res
        res_place.parse_place(place_name=res, geo_files=TestScoring.geodata.geo_files)
        res_place.feature = feat
        if res_place.country_name == '' and res_place.country_iso != '':
            res_place.country_name = TestScoring.geodata.geo_files.geodb.get_country_name(res_place.country_iso)

        score = TestScoring.scoring.match_score(in_place, res_place)
        print(f'{score:.1f} [{in_place.original_entry.title()}] [{res_place.get_five_part_title()}]')
        return score

    # palo alto, santa clara, california, usa

    def test_scr222(self):
        score = self.run_test3()
        self.assertLess(score, TestScoring.test_values[TestScoring.test_idx][3],
                        TestScoring.test_values[TestScoring.test_idx][0])

    def test_scr223(self):
        score = self.run_test3()
        self.assertLess(score, TestScoring.test_values[TestScoring.test_idx][3],
                        TestScoring.test_values[TestScoring.test_idx][0])

"""
    def test_scr22(self):
        title = "score1"
        score = self.run_test3(title, "chelsea,,england", "winchelsea, east sussex, england, united kingdom", 'PP1M')
        self.assertLess(score, TestScoring.STRONG, title)
        self.assertGreater(score, TestScoring.VERY_STRONG, title)

    def test_scr23(self):
        title = "score1"
        score = self.run_test3(title, "chelsea,,england", "chelsea, greater london, england, united kingdom", 'PP1M')
        self.assertLess(score, TestScoring.STRONG, title)
        self.assertGreater(score, TestScoring.PERFECT, title)

    def test_scr21(self):
        title = "score1"
        score = self.run_test3(title, "sonderburg", "sonderburg,sonderborg kommune,region syddanmark, denmark", 'PP1M')
        self.assertLess(score, TestScoring.GOOD, title)
        self.assertGreater(score, TestScoring.STRONG, title)

    def test_scr01(self):
        title = "score1"
        score = self.run_test3(title, "Paris, France", "Paris, France", 'PP1M')
        self.assertLess(score, TestScoring.VERY_STRONG, title)
        self.assertGreater(score, TestScoring.PERFECT, title)

    def test_scr00(self):
        title = "score0"
        score = self.run_test3(title, "Paris, France.", "Paris, France", 'PP1M')
        self.assertLess(score, TestScoring.VERY_STRONG, title)
        self.assertGreater(score, TestScoring.PERFECT, title)

    def test_scr02(self):
        title = "score2"
        score = self.run_test3(title, "London, England", "London, England, United Kingdom", 'PP1M')
        self.assertLess(score, TestScoring.VERY_STRONG, title)
        self.assertGreater(score, TestScoring.PERFECT, title)

    def test_scr72(self):
        title = "score72"
        score = self.run_test3(title, "London, England, United Kingdom", "London, England, United Kingdom", 'PP1M')
        self.assertLess(score, TestScoring.VERY_STRONG, title)
        self.assertGreater(score, TestScoring.PERFECT, title)

    def test_scr73(self):
        title = "score72"
        score = self.run_test3(title, "London, England, United Kingdom", "London, England, United Kingdom", 'HSP')
        self.assertLess(score, TestScoring.STRONG, title)
        self.assertGreater(score, TestScoring.PERFECT, title)

    def test_scr82(self):
        title = "score82"
        score = self.run_test3(title, "Domfront, Normandy", "Domfront-En-Champagne, Sarthe, Pays De La Loire, France", 'PP1M')
        self.assertLess(score, TestScoring.POOR, title)
        self.assertGreater(score, TestScoring.GOOD, title)

    def test_scr84(self):
        title = "score84"
        score = self.run_test3(title, "Domfront, Normandy", "Domfront, Department De L'Orne, Normandie, France", 'PP1M')
        self.assertLess(score, TestScoring.STRONG, title)
        self.assertGreater(score, TestScoring.PERFECT, title)

    def test_scr83(self):
        title = "score83"
        score = self.run_test3(title, "St Quentin, Aisne, Picardy, France", "St Quentin, Departement De L'Aisne, Hauts De France, France", 'PP1M')
        self.assertLess(score, TestScoring.VERY_STRONG, title)
        self.assertGreater(score, TestScoring.PERFECT, title)

    def test_scr85(self):
        title = "score85"
        score = self.run_test3(title, "Old Bond Street, London, Middlesex, England",
                               " , London, Greater London, England, United Kingdom", 'PP1M')
        self.assertLess(score, TestScoring.STRONG, title)
        self.assertGreater(score, TestScoring.VERY_STRONG, title)

    def test_scr86(self):
        title = "score86"
        score = self.run_test3(title, "Old Bond Street, London, Middlesex, England",
                               " , Museum Of London, Greater London, England, United Kingdom", 'PPL')
        self.assertLess(score, TestScoring.GOOD, title)
        self.assertGreater(score, TestScoring.VERY_STRONG, title)

    def test_scr61(self):
        title = "score61"
        score = self.run_test3(title, "zxq, xyzzy",
                               " , London, Greater London, England, United Kingdom", ' ')
        self.assertLess(score, TestScoring.NO_MATCH, title)
        self.assertGreater(score, TestScoring.VERY_POOR, title)

    def test_scr62(self):
        title = "score62"
        score = self.run_test3(title, "France",
                               "France", 'ADM0')
        self.assertLess(score, TestScoring.VERY_STRONG, title)
        self.assertGreater(score, TestScoring.PERFECT, title)

    def test_scr03(self):
        title = "score3"
        score = self.run_test3(title, "St. Margaret, Westminster, London, England", "London,England,United Kingdom", 'PPL')
        self.assertLess(score, TestScoring.POOR, title)
        self.assertGreater(score, TestScoring.GOOD, title)

    def test_scr04(self):
        title = "score4"
        score = self.run_test3(title, "St. Margaret, Westminster, London, England", "Westminster Cathedral, Greater London, England", 'PPL')
        self.assertLess(score, TestScoring.STRONG, title)
        self.assertGreater(score, TestScoring.VERY_STRONG, title)

    def test_scr05(self):
        title = "score5"
        score = self.run_test3(title, "Canada", "Canada", 'ADM0')
        self.assertLess(score, TestScoring.VERY_STRONG, title)
        self.assertGreater(score, TestScoring.PERFECT, title)

    # St Margaret, Westminster Cathedral, Greater London, England, United Kingdom

    # ===== TEST INPUT WORD REMOVAL

    def test_in01(self):
        title = "Input word1"
        out, inp = self.run_test1(title, "France", "Paris, France")
        self.assertEqual('', inp, title)

    def test_in02(self):
        title = "Input word2"
        out, inp = self.run_test1(title, "Westchester County, New York, USA", "Westchester, New York, USA")
        self.assertEqual('County,,', inp, title)

    def test_in03(self):
        title = "Input word1"
        out, inp = self.run_test1(title, "Frunce", "Paris, France")
        self.assertEqual('u', inp, title)

    def test_in04(self):
        title = "Input word1"
        out, inp = self.run_test1(title, "London,England,", "London,England,United Kingdom")
        self.assertEqual(',,', inp, title)

    def test_in05(self):
        title = "Input word1"
        out, inp = self.run_test1(title, "St. Margaret, Westminster, London, England", "London,England,United Kingdom")
        self.assertEqual('St. Margaret, Westmsr, ,', inp, title)

    def test_in06(self):
        title = "Input word1"
        out, inp = self.run_test1(title, "St. Margaret, Westminster, London, England", "Westminster Cathedral, Greater London, England")
        self.assertEqual('St. Marg, ,,', inp, title)

    # ===== TEST OUTPUT WORD REMOVAL

    def test_out01(self):
        title = "output word1"
        out, inp = self.run_test1(title, "France", "Paris, France")
        self.assertEqual('Paris,', out, title)

    def test_out02(self):
        title = "output word2"
        out, inp = self.run_test1(title, "Westchester County, New York, USA", "Westchester, New York, USA")
        self.assertEqual(',,', out, title)

    def test_out03(self):
        title = "output word1"
        out, inp = self.run_test1(title, "Frunce", "Paris, France")
        self.assertEqual('Paris, a', out, title)

    # ===== TEST total REMOVAL
    def test_tot01(self):
        title = "total word1"
        out, inp = self.run_test1(title, "France", "Paris, France")
        out, inp = self.run_test2(title, inp, out)
        self.assertEqual('', inp, title)

    def test_tot02(self):
        title = "total word1"
        out, inp = self.run_test1(title, "Frunce", "Paris, France")
        out, inp = self.run_test2(title, inp, out)
        self.assertEqual('u', inp, title)

    def test_tot03(self):
        title = "total word2"
        out, inp = self.run_test1(title, "Westchester County, New York, USA", "Westchester, New York, USA")
        out, inp = self.run_test2(title, inp, out)
        self.assertEqual('County,,', inp, title)

    # ===== TEST INPUT CHAR REMOVAL
    def test_char01(self):
        title = "Input word1"
        out, inp = self.run_test2(title, "France", "Paris, France")
        self.assertEqual('', inp, title)

    def test_char02(self):
        title = "Input word2"
        out, inp = self.run_test2(title, "Westchester County, New York, USA", "Westchester, New York, USA")
        self.assertEqual('County,,', inp, title)

    def test_char03(self):
        title = "Input word3"
        out, inp = self.run_test2(title, "Frunce", "Paris, France")
        self.assertEqual('u', inp, title)

    
    """


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import argparse
import logging
import re
from typing import List, Tuple

from geofinder import GeoKeys
from geofinder.ArgumentParserNoExit import ArgumentParserNoExit

#default_country = 'nederland'


# What type of entity is this place?
class PlaceType:
    COUNTRY = 0
    ADMIN1 = 1
    ADMIN2 = 2
    CITY = 3
    PREFIX = 4
    ADVANCED_SEARCH = 5


place_type_name_dict = {
    PlaceType.COUNTRY: 'Country',
    PlaceType.ADMIN1: 'STATE/PROVINCE',
    PlaceType.ADMIN2: 'COUNTY',
    PlaceType.CITY: ' ',
    PlaceType.ADVANCED_SEARCH: ' '
}


class Loc:
    """
    Holds the details about a Location: Name, county, state/province, country, lat/long as well as lookup result details
    Parses a name into Loc items (county, state, etc)
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.clear()
        self.event_year: int = 0

    def clear(self):
        # Place geo info
        self.original_entry: str = ""
        self.formatted_name: str = ''
        self.lat: float = float('NaN')  # Latitude
        self.lon: float = float('NaN')  # Longitude
        self.country_iso: str = ""  # Country ISO code
        self.country_name: str = ''
        self.city1: str = ""  # City or entity name
        self.admin1_name: str = ""  # Admin1 (State/province/etc)
        self.admin2_name: str = ""  # Admin2 (county)
        self.admin1_id: str = ""  # Admin1 Geoname ID
        self.admin2_id = ""  # Admin2 Geoname ID
        self.prefix: str = ""  # Prefix (entries before city)
        self.extra: str = ''  # Extra tokens that dont get put in prefix
        self.feature: str = ''  # Geoname feature code
        self.place_type: int = PlaceType.COUNTRY  # Is this a Country , Admin1 ,admin2 or city?
        self.target: str = ''  # Target for lookup
        self.geoid: str = ''
        self.prefix_commas: str = ''
        self.id = ''
        self.enclosed_by = ''
        self.standard_parse = True

        # Lookup result info
        self.status: str = ""
        self.status_detail: str = ""
        self.result_type: int = GeoKeys.Result.NO_MATCH  # Result type of lookup
        self.result_type_text: str = ''  # Text version of result type
        self.georow_list: List[Tuple] = [()]  # List of items that matched this location

        self.georow_list.clear()

    def filter(self, place_name, geo_files):
        # Advanced search parameters
        # Separate out arguments
        tokens = place_name.split(",")
        args = []
        for tkn in tokens:
            if '--' in tkn:
                args.append(tkn.strip(' '))

        # Parse options in place name
        parser = ArgumentParserNoExit(description="Parses command.")
        parser.add_argument("-f", "--feature", help=argparse.SUPPRESS)
        parser.add_argument("-i", "--iso", help=argparse.SUPPRESS)
        parser.add_argument("-c", "--country", help=argparse.SUPPRESS)
        try:
            options = parser.parse_args(args)
            self.city1 = GeoKeys.search_normalize(tokens[0], self.country_iso)
            self.target = self.city1
            if options.iso:
                self.country_iso = options.iso.lower()
            if options.country:
                self.country_iso = options.country.lower()
            if options.feature:
                self.feature = options.feature.upper()
            self.place_type = PlaceType.ADVANCED_SEARCH
        except Exception as e:
            self.logger.debug(e)
        self.logger.debug(f'ADV SEARCH: targ={self.city1} iso={self.country_iso} feat={self.feature} typ={self.place_type}')

    def parse_place(self, place_name: str, geo_files):
        """
        Given a comma separated place name, parse into its city, AdminID, country_iso and type of entity (city, country etc)
        Expected format: prefix,city,admin2,admin1,country
        self.status has Result status code
        """
        self.clear()
        self.original_entry = place_name

        # Convert open-brace and open-paren to comma.  close brace/paren will be stripped by normalize()
        res = re.sub('\[', ',', place_name)
        res = re.sub('\(', ',', res)

        tokens = res.split(",")
        token_count = len(tokens)
        self.place_type = PlaceType.CITY

        # Parse City, Admin2, Admin2, Country scanning from the right.  When there are more tokens, we capture more fields
        # Place type is the leftmost item we found - either City, Admin2, Admin2, or Country
        # self.logger.debug(f'***** PLACE [{place_name}] *****')

        if '--' in place_name:
            # Pull out filter flags if present
            self.logger.debug('filter')
            self.filter(place_name, geo_files)
            return
        elif token_count > 0:
            #  COUNTRY - right-most token should be country
            #  Format: Country
            self.place_type = PlaceType.COUNTRY
            self.country_name = GeoKeys.search_normalize(tokens[-1], "")
            self.target = self.country_name

            # Validate country
            self.country_iso = geo_files.geodb.get_country_iso(self)  # Get Country country_iso
            if self.country_iso != '':
                # self.logger.debug(f'Found country. iso = [{self.country_iso}]')
                pass
            else:
                # Last token is not COUNTRY.
                # Append blank to token list so we now have xx,admin1, blank_country
                tokens.append('')
                token_count = len(tokens)
                self.result_type = GeoKeys.Result.NO_COUNTRY
                self.country_iso = ''
                self.country_name = ''

        if token_count > 1:
            #  Format: Admin1, Country.
            #  Admin1 is 2nd to last token
            self.admin1_name = GeoKeys.search_normalize(tokens[-2], self.country_iso)
            self.admin1_name = GeoKeys.admin1_normalize(self.admin1_name, self.country_iso)

            if len(self.admin1_name) > 0:
                self.place_type = PlaceType.ADMIN1
                self.target = self.admin1_name
                # Lookup Admin1
                geo_files.geodb.get_admin1_id(self)
                if self.admin1_id != '':
                    # self.logger.debug(f'Found admin1 {self.admin1_name}')
                    pass
                else:
                    # Last token is not Admin1 - append blank
                    self.admin1_name = ''
                    # Append blank token for admin1 position
                    tokens.append('')
                    token_count = len(tokens)

        if token_count == 3 and self.admin1_name == '' and self.country_name == '':
            # Just one valid token, so take as city
            self.city1 = GeoKeys.search_normalize(tokens[-3], self.country_iso)

            if len(self.city1) > 0:
                self.place_type = PlaceType.CITY
                self.target = self.city1
        elif token_count > 2:
            #  Format: Admin2, Admin1, Country
            #  Admin2 is 3rd to last.  Note -  if Admin2 isnt found, it will look it up as city
            self.admin2_name = GeoKeys.search_normalize(tokens[-3], self.country_iso)
            self.admin2_name, modif = GeoKeys.admin2_normalize(self.admin2_name, self.country_iso)

            if len(self.admin2_name) > 0:
                self.place_type = PlaceType.ADMIN2
                self.target = self.admin2_name

        if token_count > 3:
            # Format: Prefix, City, Admin2, Admin1, Country
            # City is 4th to last token
            # Other tokens go into Prefix
            self.city1 = GeoKeys.search_normalize(tokens[-4], self.country_iso)
            if len(self.city1) > 0:
                self.place_type = PlaceType.CITY
                self.target = self.city1

            # Assign remaining tokens (if any) to prefix.  Zero'th token to 4th from end.
            for item in tokens[0:-4]:
                if len(self.prefix) > 0:
                    self.prefix += ' '
                self.prefix += str(item.strip(' '))

        # Special case for New York, New York which normally refers to the City, not county
        if self.admin2_name == 'new york' and self.place_type == PlaceType.ADMIN2:
            self.admin2_name = 'new york city'
            self.target = self.admin2_name

        self.prefix = self.prefix.strip(',')

        self.logger.debug(f"    ======= PARSE: {place_name} City [{self.city1}] Adm2 [{self.admin2_name}]"
                          f" Adm1 [{self.admin1_name}] adm1_id [{self.admin1_id}] Cntry [{self.country_name}] Pref=[{self.prefix}]"
                          f" type_id={self.place_type}")
        return

    def get_status(self) -> str:
        self.logger.debug(f'status=[{self.status}]')
        return self.status

    def safe_strings(self):
        self.city1 = str(self.city1)
        self.admin1_name = str(self.admin1_name)
        self.admin2_name = str(self.admin2_name)
        self.prefix = str(self.prefix)

    def format_full_nm(self, replace_dct):
        """ Take the parts of a Place and build fullname.  e.g. pref, city,adm2,adm1,country name """
        self.set_place_type()

        if self.admin1_name is None:
            self.admin1_name = ''
        if self.admin2_name is None:
            self.admin2_name = ''

        if self.place_type == PlaceType.COUNTRY:
            nm = f"{self.country_name}"
        elif self.place_type == PlaceType.ADMIN1:
            nm = f"{self.admin1_name}, {self.country_name}"
        elif self.place_type == PlaceType.ADMIN2:
            nm = f"{self.admin2_name}, {self.admin1_name}, {self.country_name}"
        else:
            nm = f"{self.city1}, {self.admin2_name}, {self.admin1_name}, {str(self.country_name)}"

        if self.prefix in nm:
            self.prefix = ''

        if len(self.prefix) > 0:
            self.prefix_commas = ', '
        else:
            self.prefix_commas = ''

        nm = GeoKeys.capwords(nm)

        # Perform any text replacements user entered into Output Tab
        if replace_dct:
            for key in replace_dct:
                nm = re.sub(key, replace_dct[key], nm)

        return nm

    def get_five_part_title(self):
        # Returns a five part title string and tokenized version:
        #     prefix,city,county,state,country

        # Force type to City to generate four part title (then we add prefix for five parts)
        save_type = self.place_type
        self.place_type = PlaceType.CITY

        # Normalize country name
        save_country = self.country_name
        self.country_name, modified = GeoKeys.country_normalize(self.country_name)

        if len(self.extra) > 0:
            full_title = self.prefix + ' ' + self.extra + ',' + self.format_full_nm(None)
        else:
            full_title = self.prefix + ',' + self.format_full_nm(None)

        # Restore values to original
        self.place_type = save_type
        self.country_name = save_country

        return full_title

    def feature_to_type(self):
        # Set place type based on DB response feature code
        if self.feature == 'ADM0':
            self.place_type = PlaceType.COUNTRY
        elif self.feature == 'ADM1':
            self.place_type = PlaceType.ADMIN1
        elif self.feature == 'ADM2':
            self.place_type = PlaceType.ADMIN2
        else:
            self.place_type = PlaceType.CITY
        if len(self.prefix) > 0:
            self.place_type = PlaceType.PREFIX
        return self.place_type

    def set_place_type(self):
        # Set place type based on parsing results
        self.place_type = PlaceType.CITY
        if len(str(self.country_name)) > 0:
            self.place_type = PlaceType.COUNTRY
        if len(self.admin1_name) > 0:
            self.place_type = PlaceType.ADMIN1
        if len(self.admin2_name) > 0:
            self.place_type = PlaceType.ADMIN2
        if len(self.city1) > 0:
            self.place_type = PlaceType.CITY

    def remove_old_fields(self):
        if self.place_type == PlaceType.COUNTRY:
            self.prefix = ''
            self.city1 = ''
            self.admin2_name = ''
            self.admin1_name = ''
        elif self.place_type == PlaceType.ADMIN1:
            self.prefix = ''
            self.city1 = ''
            self.admin2_name = ''
        elif self.place_type == PlaceType.ADMIN2:
            self.prefix = ''
            self.city1 = ''
        elif self.place_type == PlaceType.CITY:
            self.prefix = ''

    def set_place_type_text(self):
        if self.result_type == GeoKeys.Result.NO_COUNTRY:
            self.result_type_text = 'Country'
        elif self.place_type == PlaceType.COUNTRY:
            self.result_type_text = 'Country'
        elif self.place_type == PlaceType.ADMIN1:
            self.result_type_text = self.get_district1_type(self.country_iso)
        elif self.place_type == PlaceType.ADMIN2:
            self.result_type_text = 'County'
        elif self.place_type == PlaceType.CITY:
            self.result_type_text = self.get_type_name(self.feature)
        elif self.place_type == PlaceType.PREFIX:
            self.result_type_text = 'Place'

    @staticmethod
    def get_district1_type(iso) -> str:
        # Return the local country term for Admin1 district
        if iso in ["al", "no"]:
            return "County"
        elif iso in ["us", "at", "bm", "br", "de"]:
            return "State"
        elif iso in ["ac", "an", 'ao', 'bb', 'bd']:
            return "Parish"
        elif iso in ["ae"]:
            return "Emirate"
        elif iso in ["bc", "bf", "bh", "bl", "bn"]:
            return "District"
        elif iso in ["gb"]:
            return "Country"
        else:
            return "Province"

    @staticmethod
    def get_type_name(feature):
        nm = type_name.get(feature)
        if nm is None:
            nm = ''
        return nm


type_name = {"ADM0": 'Country', "ADM1": 'City', "ADM2": 'City', "ADM3": 'City', "ADM4": 'City', "ADMF": 'City',
             "CH": 'Church', "CSTL": 'Castle', "CMTY": 'Cemetery', "EST": 'Estate', "HSP": 'Hospital',
             "HSTS": 'Historic', "ISL": 'Island', "MSQE": 'Mosque', "MSTY": 'Monastery', "MT": 'Mountain', "MUS": 'Museum', "PAL": 'Palace',
             "PPL": 'City', "PPLA": 'City', "PPLA2": 'City', "PPLA3": 'City', "PPLA4": 'City',
             "PPLC": 'City', "PPLG": 'City', "PPLH": 'City', "PPLL": 'Village', "PPLQ": 'City', "PPLX": 'City',
             "PRK": 'Park', "PRN": 'Prison', "PRSH": 'Parish', "RUIN": 'Ruin',
             "RLG": 'Religious', "STG": '', "SQR": 'Square', "SYG": 'Synagogue', "VAL": 'Valley', "PP1M": 'City', "PP1K": 'City'}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
from tkinter import ttk

TXT_WID = 95
BTN_WID = 5
BTN_WID_WD = 6
THIN_WID = 6

BTN_PADX = 50
PAD_PADX = 25

# Color Palette
LT_GRAY = 'gray97'
LT_COLOR = 'gray50'
FG_COLOR = 'gray29'
BG_COLOR = "white"
HIGH_COLOR = 'royalblue3'
ERR_COLOR = 'red2'
GOOD_COLOR = 'green4'
ODD_ROW_COLOR = '#E8E8f8'

# Font
FNT_NAME = 'Helvetica'


FNT_SIZE_SM = 10
FNT_SIZE_MD = 12
FNT_SIZE_LG = 14
FNT_SIZE_XL = 16


class GFStyle:
    """ Tk Styles for GeoFinder and GeoUtil """

    def __init__(self):
        styl = ttk.Style()

        styl.theme_use("clam")
        styl.configure('.', font=(FNT_NAME, FNT_SIZE_MD))  # Default button font

        # Label colors
        styl.configure('Large.TLabel', foreground=HIGH_COLOR, background=BG_COLOR, font=(FNT_NAME, FNT_SIZE_XL))

        styl.configure('Error.TLabel', foreground=ERR_COLOR, background=BG_COLOR, font=(FNT_NAME, FNT_SIZE_LG))
        styl.configure('Good.TLabel', foreground=GOOD_COLOR, background=BG_COLOR, font=(FNT_NAME, FNT_SIZE_LG))
        styl.configure('GoodCounty.TLabel', foreground=HIGH_COLOR, background=BG_COLOR, font=(FNT_NAME, FNT_SIZE_LG))

        styl.configure('Info.TLabel', foreground=FG_COLOR, background=BG_COLOR, font=(FNT_NAME, FNT_SIZE_MD))
        styl.configure('Highlight.TLabel', foreground=HIGH_COLOR, background=BG_COLOR, font=(FNT_NAME, FNT_SIZE_MD))
        styl.configure('Light.TLabel', borderwidth=3, foreground=LT_COLOR, background=BG_COLOR,
                       font=(FNT_NAME, FNT_SIZE_MD))

        styl.configure('Tiny.TLabel', borderwidth=3, foreground=LT_COLOR, background=BG_COLOR,
                       font=(FNT_NAME, FNT_SIZE_SM))

        # Entry field and button colors
        styl.configure('TEntry', foreground=FG_COLOR)
        styl.configure('TButton', foreground=FG_COLOR)
        styl.configure('Preferred.TButton', foreground=GOOD_COLOR)

        # Treeview (listbox) colors
        styl.configure("Plain.Treeview", highlightthickness=0, bd=0,rowheight=25, font=(FNT_NAME, FNT_SIZE_MD),
                       foreground=FG_COLOR)  # Modify the font of the body
        styl.configure("Plain.Treeview.Heading", font=(FNT_NAME, FNT_SIZE_MD, 'bold'),
                       foreground=FG_COLOR, background=LT_GRAY,rowheight=30, relief="flat")  # Modify the font of
        # treeview heading
        styl.layout("Plain.Treeview", [('Plain.Treeview.treearea', {'sticky': 'nswe'})])  # Remove the borders

        # Notebook colors
        styl.element_create('Plain.Notebook.tab', "from", 'default')
        # Redefine the TNotebook Tab layout to use the new element
        styl.layout("TNotebook.Tab",
                    [('Plain.Notebook.tab', {'children':
                                                 [('Notebook.padding', {'side': 'top', 'children':
                                                     [('Notebook.focus', {'side': 'top', 'children':
                                                         [('Notebook.label', {'side': 'top', 'sticky': ''})],
                                                                          'sticky': 'nswe'})],
                                                                        'sticky': 'nswe'})],
                                             'sticky': 'nswe'})])
        styl.configure("TNotebook", background=BG_COLOR, borderwidth=0)
        styl.configure("TNotebook.Tab", background="purple", foreground=FG_COLOR,
                       lightcolor='red', borderwidth=0)

        # Frame color
        styl.configure("TFrame", background=BG_COLOR, foreground=FG_COLOR, borderwidth=0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# See https://setuptools.readthedocs.io/en/latest/setuptools.html
# To use the 'upload' functionality of this file:
# This will create a PyPi package and publish it.  It will also tag the version in github
#   python3 Setup.py upload   (increment __version__.py first)

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = "geofinder"
DESCRIPTION = "GEDCOM Genealogy address validation and geocoding using geonames.org data"
URL = "https://github.com/corb555/GeoFinder"
EMAIL = "corb@aol.com"
AUTHOR = "Mike Herbert"
REQUIRES_PYTHON = '>=3.6.0'
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    'unidecode', 'phonetics'
]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

print('Python {}.{}'.format(sys.version_info[0], sys.version_info[1]))
if sys.version_info < (3, 6, 0):
    raise Exception("Must be using Python 3.6 or higher")

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, '__version__.py')) as f:
        exec(f.read(), about)
    tkn = about['__version__'].split('.')
    next_build = int(tkn[-1]) + 1
    # next_ver = f'{tkn[0]}.{tkn[1]}.{next_build}'
    # print(next_ver)
else:
    about['__version__'] = VERSION


class UploadCommand(Command):
    """Support setup.py upload - create release version tag in Github.  uploads package to PyPi"""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('python3 setup.py sdist bdist_wheel')

        self.status('Uploading the package to PyPI via Twine…')
        os.system('python3 -m twine upload  dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{}'.format(about["__version__"]))
        os.system('git push --tags')

        sys.exit()


# Setup:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],

    entry_points={
        'console_scripts': [
            'geofinder = geofinder.GeoFinder:entry'
        ],
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    package_data={'geofinder': ['images/*.gif']},
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "Natural Language :: English",
        "Topic :: Sociology :: Genealogy"
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)



import string as st
import re

text =["'s gemeenthe", "st paul's cathedral", "washington, d.c."]

for item in text:
    nm = st.capwords(item)
    print(nm)
    nm = re.sub(r"(?:[a-z])(\'S )","'s ", nm)
    nm = re.sub(r"D.c.", "D.C.", nm)
    print (nm)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import logging
import re
from difflib import SequenceMatcher

from geofinder import GeoKeys, Geodata, Loc


class MatchScore:
    """
    Calculate how close the database result place name is to the users input place name.
    1) Recursively remove largest text sequence in both to end up with just mismatched text
    2) Calculate the percent that didnt match in each comma separated term of user input
    3) Score is based on percent mismatch weighted for each term (City is higher, county is lower)

    A standard text difference, such as Levenstein, was not used because those treat both strings as equal, whereas this
    treats the User text as more important than DB result and also weights each token

    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.weight = [0.7, 1.0, 0.2, 0.6, 0.9]

        # Out weight + Feature weight must be less than 1.0.
        self.out_weight = 0.17
        self.feature_weight = 0.06
        if self.out_weight + self.feature_weight > 1.0:
            self.logger.error('Out weight + Feature weight must be less than 1.0')

        self.wildcard_penalty = 20.0
        self.first_token_match_bonus = 27.0
        self.wrong_order_penalty = 2.0

    def match_score(self, inp_place: Loc.Loc, res_place: Loc.Loc) -> int:
        """
        :param inp_place: Input place structure with users text
        :param res_place: Result place structure with DB result
        :return: score 0-100 reflecting the difference between the user input and the result.  0 is perfect match, 100 is no match
        Score is also adjusted based on Feature type.  More important features (large city) get lower result
        """
        inp_len = [0] * 5
        num_inp_tokens = 0.0
        in_score = 0

        # Create full place title (prefix,city,county,state,country) from input place.
        inp_title = inp_place.get_five_part_title()
        inp_title = GeoKeys.normalize_match_title(inp_title, inp_place.country_iso)
        inp_tokens = inp_title.split(',')

        # Create full place title (prefix,city,county,state,country) from result place
        res_place.prefix = ' '
        res_title = res_place.get_five_part_title()
        res_title = GeoKeys.normalize_match_title(res_title, res_place.country_iso)
        res_tokens = res_title.split(',')

        # Store length of original input tokens.  This is used for percent unmatched calculation
        for it, tk in enumerate(inp_tokens):
            inp_tokens[it] = inp_tokens[it].strip(' ')
            inp_len[it] = len(inp_tokens[it])

        # Create a list of all the words in result and save result len for percent calc
        res_word_list = ', '.join(map(str, res_tokens))
        orig_res_len = len(res_word_list)

        # Create a list of all the words in input
        input_words = ', '.join(map(str, inp_tokens))

        # Remove any matching sequences in input list and result
        res_word_list, input_words = self.remove_matching_sequences(res_word_list, input_words)

        # For each input token calculate percent of new (unmatched) size vs original size
        unmatched_input_tokens = input_words.split(',')

        # Each token in place hierarchy gets a different weighting
        #      Prefix, city,county, state, country
        score_diags = ''

        # Calculate percent of USER INPUT text that was unmatched, then apply weighting
        for idx, tk in enumerate(inp_tokens):
            if inp_len[idx] > 0:
                unmatched_percent = int(100.0 * len(unmatched_input_tokens[idx].strip(' ')) / inp_len[idx])
                in_score += unmatched_percent * self.weight[idx]
                score_diags += f'  {idx}) [{tk}]{inp_len[idx]} {unmatched_percent}% * {self.weight[idx]} '
                # self.logger.debug(f'{idx}) Rem=[{unmatched_input_tokens[idx].strip(" " )}] wgtd={unmatched_percent * self.weight[idx]}')
                num_inp_tokens += 1.0 * self.weight[idx]
                # self.logger.debug(f'{idx} [{inp_tokens2[idx]}:{inp_tokens[idx]}] rawscr={sc}% orig_len={inp_len[idx]} wgt={self.weight[idx]}')
                if idx < 2:
                    # If the full first or second token of the result is in input then improve score
                    # Bonus for a full match as against above partial matches
                    if res_tokens[idx] in inp_tokens[idx]:
                        in_score -= self.first_token_match_bonus

        # Average over number of tokens (with fractional weight).  Gives 0-100% regardless of weighting and number of tokens
        in_score = in_score / num_inp_tokens
        # self.logger.debug(f'raw in={in_score}  numtkn={num_inp_tokens}')

        # Calculate percent of DB RESULT text that was unmatched
        if orig_res_len > 0:
            out_score = int(100.0 * len(res_word_list.strip(' ')) / orig_res_len)
            # self.logger.debug(f"Out=[{res_word_list.strip(' ')}] orig_len={orig_res_len}")
        else:
            out_score = 0

        if not inp_place.standard_parse:
            # If Tokens were not in hierarchical order, give penalty
            parse_penalty = self.wrong_order_penalty
        else:
            parse_penalty = 0.0

        if '*' in inp_place.original_entry:
            # if it was a wildcard search it's hard to rank - add a penalty
            wildcard_penalty = self.wildcard_penalty
        else:
            wildcard_penalty = 0.0

        # Feature score is to ensure "important" places  get  higher rank (large city, etc)
        feature_score = Geodata.Geodata.get_priority(res_place.feature)

        # Add up scores - Each item is 0-100 and weighed as below
        in_weight = 1.0 - self.out_weight - self.feature_weight

        score = in_score * in_weight +  out_score * self.out_weight  + feature_score * self.feature_weight + parse_penalty + wildcard_penalty

        # self.logger.debug(f'SCORE {score:.1f} [{res_title}]  out={out_score * out_weight:.1f} '
        #                  f'in={in_score:.1f} feat={feature_score * feature_weight:.1f} parse={parse_penalty}\n {score_diags}')

        return score

    def _remove_matching_seq(self, text1: str, text2: str, attempts: int) -> (str, str):
        """
        Find largest matching sequence.  Remove it in text1 and text2.
                Private - called by remove_matching_sequences which provides a wrapper
        Call recursively until attempts hits zero or there are no matches longer than 1 char
        :param text1:
        :param text2:
        :param attempts: Number of times to remove largest text sequence
        :return:
        """
        s = SequenceMatcher(None, text1, text2)
        match = s.find_longest_match(0, len(text1), 0, len(text2))
        if match.size > 1:
            # Remove matched sequence from inp and out
            item = text1[match.a:match.a + match.size]
            text2 = re.sub(item, '', text2)
            text1 = re.sub(item, '', text1)
            if attempts > 0:
                # Call recursively to get next largest match and remove it
                text1, text2 = self._remove_matching_seq(text1, text2, attempts - 1)
        return text1, text2

    def remove_matching_sequences(self, text1: str, text2: str) -> (str, str):
        """
        Find largest sequences that match between text1 and 2.  Remove them from text1 and text2.
        :param text1:
        :param text2:
        :return:
        """
        # Prepare strings for input to remove_matching_seq
        # Swap all commas in text1 string to '@'.  This way they will never match comma in text2 string
        # Ensures we don;t remove commas and don't match across tokens
        text2 = re.sub(',', '@', text2)
        text1, text2 = self._remove_matching_seq(text1=text1, text2=text2, attempts=15)
        # Restore commas in inp
        text2 = re.sub('@', ',', text2)
        return text1.strip(' '), text2.strip(' ')
